
/* ICFP 2001 entry by Malcolm Parsons
 * licensed under GNU GPL version 2 or later.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define false 0
#define true 1 

#define BUFSIZE 2000

typedef char bool;

typedef struct buflist_s
{
    char buf[BUFSIZE];
    struct buflist_s *next;
} buflist_t;

typedef struct property_s
{
    bool b;
    bool em;
    bool i;
    bool s;
    bool tt;
    char u;
    char size;
    char color;
    char tag[3];
    struct property_s *previous;
} property_t;

typedef struct dchar_s
{
    char *text;
    property_t *property;
    int bcount;
    int emcount;
    int icount;
    int scount;
    int ttcount;
    int ucount;
    int sizecount;
    int colorcount;
    struct dchar_s *next;
    struct dchar_s *previous;
} dchar_t;

buflist_t *inbuf = NULL, *inbufhead = NULL;
int inbuflen = 0;
buflist_t *outbuf = NULL, *outbufhead = NULL;
int outbuflen = 0;

property_t *initProperty(void)
{
    property_t *p;

    p = (property_t *) calloc(1, sizeof(property_t));
    p->b = p->em = p->i = p->s = p->tt = false;
    p->u = 0;
    p->size = -1;
    p->color = 'i';
    p->previous = NULL;

    return p;
}

property_t *newProperty(property_t *p)
{
    property_t *q;

    q = (property_t *) calloc(1, sizeof(property_t));
    q->b = p->b;
    q->em = p->em;
    q->i = p->i;
    q->s = p->s;
    q->tt = p->tt;
    q->u = p->u;
    q->size = p->size;
    q->color = p->color;
    q->previous = p;

    return q;
}

dchar_t *initDChar(property_t *p)
{
    dchar_t *d;

    d = (dchar_t *) calloc(1, sizeof(dchar_t));
    d->text = NULL;
    d->property = p;
    d->next = NULL;
    d->previous = NULL;

    return d;
}

dchar_t *newDChar(dchar_t *d, property_t *p, char text[81], int *i)
{
    text[*i]    = '\0';
    d->next     = (dchar_t *) calloc(1, sizeof(dchar_t));
    d->next->previous = d;
    d           = d->next;
    d->text     = strdup(text);
    d->property = p;
    d->next     = NULL;
    *i          = 0;

    return d;
}

int nextChar()
{
    static int i = 0;
    int c;

    if (NULL == inbufhead)
    {
        inbufhead = (buflist_t *)calloc(1,sizeof(buflist_t));
        inbuf = inbufhead;
    }

    if (i == 80)
    {
        i = 0;
        inbuf->next = (buflist_t *)calloc(1,sizeof(buflist_t));
        inbuf = inbuf->next;
    }

    c = getchar();
    if (EOF != c)
    {
        inbuf->buf[i++] = c;
        inbuflen++;
    }

    return c;
}

void addOutput(char *s)
{
    static int i;
    int j;

    if (NULL == outbufhead)
    {
        outbufhead = (buflist_t *)calloc(1,sizeof(buflist_t));
        outbuf = outbufhead;
    }

    j = strlen(s);
    if (i + j >= BUFSIZE)
    {
        outbuf->buf[i] = '\0';
        i = 0;
        outbuf->next = (buflist_t *)calloc(1,sizeof(buflist_t));
        outbuf = outbuf->next;
    }

    
    for (j = 0; s[j] != '\0'; j++)
    {
        outbuf->buf[i++] = s[j];
    }

    outbuflen += j;
}

void printBuf(buflist_t *b)
{
    while (NULL != b)
    {
        printf("%s",b->buf);
        b = b->next;
    }
}

dchar_t *parse(void)
{
    property_t *p, *r;
    dchar_t *d, *e;
    int c,i = 0;
    char text[BUFSIZE+1], lastchar = '\0';

    p = initProperty();
    r = p;

    e = initDChar(p);
    d = e;

    while((c = nextChar()) != EOF)
    {
        if (i == BUFSIZE)
            d = newDChar(d, p, text, &i);
        switch(c)
        {
            case '<':
                if (i!=0)
                    d = newDChar(d, p, text, &i);
                c = nextChar();
                if ('/' == c)
                {
                    p = p->previous;
                    if (NULL == p)
                        exit(1);
                    while ('>' != c)
                        c = nextChar();
                    break;
                }
                p = newProperty(p);
                switch (c)
                {
                    case 'B':
                        p->b = true;
                        break;
                    case 'E':
                        nextChar();
                        if (true != p->s)
                        {
                            p->em = 1 - p->em;
                        }
                        break;
                    case 'I':
                        p->i = true;
                        break;
                    case 'P':
                        nextChar();
                        p->s = false;
                        p->em = false;
                        p->i = false;
                        p->b = false;
                        p->tt = false;
                        p->u = 0;
                        break;
                    case 'S':
                        p->s = true;
                        p->em = false;
                        break;
                    case 'T':
                        nextChar();
                        p->tt = true;
                        break;
                    case 'U':
                        if (p->u < 3)
                        {
                            p->u++;
                        }
                        break;
                    case '0':
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9':
                        p->size = c - '0';
                        break;
                    case 'r':
                    case 'g':
                    case 'b':
                    case 'c':
                    case 'm':
                    case 'y':
                    case 'k':
                    case 'w':
                        p->color = c;
                        break;
                    default:
                        exit (1);
                }
                c = nextChar();
                break;
            case '\r':
            case '\n':
            case ' ':
            case '\t':
                if (true == p->tt)
                {
                    text[i++] = c;
                    lastchar = ' ';
                    r = p;
                }
                else
                {
                    if(' ' != lastchar || 
                       p->u != r->u ||
                       (r->u != 0 &&
                       p->color != r->color) ||
                       p->size != r->size ||
                       p->tt != r->tt)
                    {
                        text[i++] = ' ';
                        r = p;
                        lastchar = ' ';
                    }
                }
                break;
            default:
                text[i++] = c;
                lastchar = c;
                break;
        }
    }

    if (i!=0)
    {
        d = newDChar(d, p, text, &i);
    }

    return e;
}

void debugoutput(dchar_t *e)
{
    dchar_t *d;

    fprintf(stderr, " len     b    em     i     s    tt      u    size  color\n");
    for (d = e->next; NULL != d; d = d->next)
    {
        fprintf(stderr, "%4d  %c%3d  %c%3d  %c%3d  %c%3d  %c%3d  %c%3d  %c%3d  %c%3d",
                strlen(d->text),
                d->property->b == true ? 'B' : ' ' ,     d->bcount,
                d->property->em == true ? 'E' : ' ',    d->emcount,
                d->property->i == true ? 'I' : ' ',     d->icount,
                d->property->s == true ? 'S' : ' ',     d->scount,
                d->property->tt == true ? 'T': ' ',    d->ttcount,
                d->property->u == 0 ? ' ': '0' + d->property->u,     d->ucount,
                d->property->size == -1 ? ' ' : '0' + d->property->size,  d->sizecount,
                d->property->color == 'i' ? ' ' : d->property->color, d->colorcount);
        fprintf(stderr, " \"%s\"\n", d->text);
    }
}

int colorindex(char color)
{
    switch (color)
    {
        case 'r':
            return 0;
        case 'g':
            return 1;
        case 'b':
            return 2;
        case 'c':
            return 3;
        case 'y':
            return 4;
        case 'm':
            return 5;
        case 'k':
            return 6;
        case 'w':
            return 7;
    }
    return 0;
}
    
void count(dchar_t *head)
{
    dchar_t *d;
    dchar_t *B,*EM,*I,*S,*TT,*U[4],*SIZE[10],*COLOR[8];
    int u = 0, i, size;
    char color;

    B = EM = I = S = TT = U[0] = head->next;
    for (i = 0; i < 10; i++)
        SIZE[i] = NULL;
    for (i = 0; i < 8; i++)
        COLOR[i] = NULL;

    for(d = head->next; NULL != d; d = d->next)
    {
        if (strlen(d->text) == 1 && d->text[0] == ' ')
        {
            if (d->property->u != 0)
            {
                if (color != d->property->color)
                {
                    color = d->property->color;
                    if ('i' == color)
                    {
                        for (i = 0; i < 8; i++)
                            COLOR[i] = NULL;
                    }
                    else
                    {
                        i = colorindex(color);
                        if (NULL == COLOR[i])
                            COLOR[i] = d;
                    }
                }
            }
        }
        else
        {
            if (B->property->b != d->property->b) B = d;
            if (EM->property->em != d->property->em && d->property->s == false) EM = d;
            if (I->property->i != d->property->i) I = d;
            if (S->property->s != d->property->s) S = d;
            if (color != d->property->color)
            {
               color = d->property->color;
               if ('i' == color)
               {
                   for (i = 0; i < 8; i++)
                       COLOR[i] = NULL;
               }
               else
               {
                   i = colorindex(color);
                   if (NULL == COLOR[i])
                       COLOR[i] = d;
               }
            }
        }
        if (size != d->property->size)
        {
            size = d->property->size;
            if (-1 == d->property->size)
            {
               for (i = 0; i < 10; i++)
                  SIZE[i] = NULL;
            }
            else
            {
                if (NULL == SIZE[size])
                    SIZE[size] = d;
            }
        }
        if (TT->property->tt != d->property->tt) TT = d;
        if (u != d->property->u) 
        {
            if (u < d->property->u)
            {
                u = d->property->u;
                U[u] = d;
            }
            else
            {
                u = d->property->u;
            }
        }

        B->bcount++;
        EM->emcount++;
        I->icount++;
        S->scount++;
        TT->ttcount++;
        for (i = 0; i <= u; i++)
        {
            U[i]->ucount++;
        }
        for (i = 0; i < 10; i++)
        {
            if (SIZE[i] != NULL)
                SIZE[i]->sizecount++;
        }
        for (i = 0; i < 8; i++)
        {
            if (COLOR[i] != NULL)
                COLOR[i]->colorcount++;
        }
    }
}

int maxCount(dchar_t *d, int limit)
{
    int max = 0;

    if (d->bcount > max && d->bcount < limit)         max = d->bcount;
    if (d->icount > max && d->icount < limit)         max = d->icount;
    if (d->scount > max && d->scount < limit)         max = d->scount;
    if (d->ttcount > max && d->ttcount < limit)       max = d->ttcount;
    if (d->emcount > max && d->emcount < limit)       max = d->emcount;
    if (d->ucount > max && d->ucount < limit)         max = d->ucount;
    if (d->sizecount > max && d->sizecount < limit)   max = d->sizecount;
    if (d->colorcount > max && d->colorcount < limit) max = d->colorcount;

    return max;
}

void opentags(dchar_t *d, property_t **p)
{
    int max = 1e4;
    char s[10];

    max = maxCount(d, max);
    while (max > 0)
    {
        if (true == d->property->b && max == d->bcount)
        {
            *p = newProperty(*p);
            (*p)->b = true;
            strcpy((*p)->tag, "B");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->i && max == d->icount)
        {
            *p = newProperty(*p);
            (*p)->i = true;
            strcpy((*p)->tag, "I");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->s && max == d->scount)
        {
            *p = newProperty(*p);
            (*p)->s = true;
            (*p)->em = false;
            strcpy((*p)->tag, "S");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->em && false == d->property->s && max == d->emcount)
        {
            *p = newProperty(*p);
            (*p)->em = true;
            strcpy((*p)->tag, "EM");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->tt && max == d->ttcount)
        {
            *p = newProperty(*p);
            (*p)->tt = true;
            strcpy((*p)->tag, "TT");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (d->property->u > (*p)->u && max == d->ucount)
        {
            while (d->property->u > (*p)->u)
            {
                *p = newProperty(*p);
                if ((*p)->u < 3)
                {
                    (*p)->u++;
                }
                strcpy((*p)->tag, "U");
                sprintf(s, "<%s>", (*p)->tag);
                addOutput(s);
            }
        }
        if (-1 != d->property->size && max == d->sizecount)
        {
            *p = newProperty(*p);
            (*p)->size = d->property->size;
            (*p)->tag[0] = '0' + d->property->size;
            (*p)->tag[1] = '\0';
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if ('i' != d->property->color && max == d->colorcount)
        {
            *p = newProperty(*p);
            (*p)->color = d->property->color;
            (*p)->tag[0] = d->property->color;
            (*p)->tag[1] = '\0';
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        max = maxCount(d, max);
    }
    if (strlen(d->text) == 1 && d->text[0] == ' ')
    {
        if (true == d->property->u)
        {
            if ('i' != d->property->color && d->property->color != (*p)->color)
            {
                *p = newProperty(*p);
                (*p)->color = d->property->color;
                (*p)->tag[0] = d->property->color;
                (*p)->tag[1] = '\0';
                sprintf(s, "<%s>", (*p)->tag);
                addOutput(s);
            }
        }

    }
    else
    {
        if (true == d->property->b && false == (*p)->b)
        {
            *p = newProperty(*p);
            (*p)->b = true;
            strcpy((*p)->tag, "B");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->i && false == (*p)->i)
        {
            *p = newProperty(*p);
            (*p)->i = true;
            strcpy((*p)->tag, "I");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->s && false == (*p)->s)
        {
            *p = newProperty(*p);
            (*p)->s = true;
            strcpy((*p)->tag, "S");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if (true == d->property->em && false == (*p)->em)
        {
            *p = newProperty(*p);
            (*p)->em = true;
            strcpy((*p)->tag, "EM");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
        if ('i' != d->property->color && d->property->color != (*p)->color)
        {
            *p = newProperty(*p);
            (*p)->color = d->property->color;
            (*p)->tag[0] = d->property->color;
            (*p)->tag[1] = '\0';
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
    }
    if (true == d->property->tt && false == (*p)->tt)
    {
        *p = newProperty(*p);
        (*p)->tt = true;
        strcpy((*p)->tag, "TT");
        sprintf(s, "<%s>", (*p)->tag);
        addOutput(s);
    }
    if (d->property->size != (*p)->size)
    {
        *p = newProperty(*p);
        (*p)->size = d->property->size;
        (*p)->tag[0] = '0' + d->property->size;
        (*p)->tag[1] = '\0';
        sprintf(s, "<%s>", (*p)->tag);
        addOutput(s);
    }
    if (d->property->u != (*p)->u)
    {
        while (d->property->u != (*p)->u)
        {
            *p = newProperty(*p);
            if ((*p)->u < 3)
            {
                (*p)->u++;
            }
            strcpy((*p)->tag, "U");
            sprintf(s, "<%s>", (*p)->tag);
            addOutput(s);
        }
    }
}

bool propertyEqual(property_t *p, property_t *q)
{
    if (p->b != q->b)         return false;
    if (p->i != q->i)         return false;
    if (p->em != q->em)       return false;
    if (p->s != q->s)         return false;
    if (p->tt != q->tt)       return false;
    if (p->u != q->u)         return false;
    if (p->size != q->size)   return false;
    if (p->color != q->color) return false;
    return true;
}

bool propertyNeedsClose(property_t *p, property_t *q, bool flag)
{
    if (false == flag)
    {
        if (p->b && !q->b)                    return true;
        if (p->i && !q->i)                    return true;
        if (p->em && !q->em)                  return true;
        if (p->s && !q->s)                    return true;
        if ('i' != p->color && 'i' == q->color) return true;
    }
    else
    {
        if (q->u != 0)
        {
            if ('i' != p->color && 'i' == q->color) return true;
        }
    }
    if (-1 != p->size && -1 == q->size)   return true;
    if (p->tt && !q->tt)                  return true;
    if (p->u > q->u)                      return true;
    return false;
}

void closetags(dchar_t *d, property_t **p)
{
    bool flag = false;
    char s[10];
    
    if (strlen(d->text) == 1 && d->text[0] == ' ')
    {
        flag = true;
    }
    while (propertyNeedsClose(*p, d->property, flag))
    {
        sprintf(s, "</%s>",(*p)->tag);
        addOutput(s);
        (*p) = (*p)->previous;
    }
    flag = true;
    while (true == flag)
    {
        flag = false;
        if (NULL != (*p)->previous)
        {
            if (d->property->color != (*p)->color && d->property->color == (*p)->previous->color)
            {
                flag = true;
                sprintf(s, "</%s>",(*p)->tag);
                addOutput(s);
                (*p) = (*p)->previous;
            }
            if (d->property->size != (*p)->size && d->property->size == (*p)->previous->size)
            {
                flag = true;
                sprintf(s, "</%s>",(*p)->tag);
                addOutput(s);
                (*p) = (*p)->previous;
            }
        }
    }
}

void closealltags(property_t *p)
{
    char s[10];
    while (NULL != p->previous)
    {
        sprintf(s, "</%s>",p->tag);
        addOutput(s);
        p = p->previous;
    }
}

void output(dchar_t *e)
{
    property_t *p;
    dchar_t *d;

    p = e->property;
    for (d = e->next; NULL != d; d = d->next)
    {
        closetags(d, &p);
        opentags(d, &p);
        addOutput(d->text);
    }
    closealltags(p);
}

int main (int argc, char *argv[])
{
    dchar_t *d;

    d = parse();

    count(d);

    if (argc > 2)
    {
        debugoutput(d);
    }

    output(d);

    if (argc > 2)
    {
        fprintf(stderr, "inputfile was %d bytes\noutputfile is %d bytes\n",inbuflen,outbuflen);
    }

    printBuf(inbuflen < outbuflen ? inbufhead : outbufhead);

    return 0;
}

